/****************************************************************************
 * O2Micro, Inc. presidio device driver for Linux.
 *
 *
 * Revisions 0.1
 * toby.yang@o2micro.com
 * 2004/07/27
 ****************************************************************************/

//#define MODULE
//#include <linux/config.h>
//#include <linux/version.h>
#include <linux/module.h>


//#if (LINUX_VERSION_CODE >= 131343)
#include <linux/init.h>
//#endif
#include "base.h"
#include "basic.h"
#include "scsi_o2media.h"
#include "scsi_o2sd.h"
/* Deal with CONFIG_MODVERSIONS */
/*
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif      
*/
/* Following function is for manully find pci device */

#define INCLUDE_O2MEDIA
#define INCLUDE_O2SD
static unsigned long  PCI_IND(unsigned long ioaddr)
{
        outl(ioaddr,PCI_PORT_ADDRESS);
        return inl(PCI_PORT_DATA);
}

static void  PCI_OUTD(unsigned long ioaddr,unsigned long data)
{

        outl(ioaddr,PCI_PORT_ADDRESS);
        outl(data,PCI_PORT_DATA);
}


static unsigned short PCI_INW(unsigned long ioaddr)
{
        outl(ioaddr,PCI_PORT_ADDRESS);
        return inw(PCI_PORT_DATA);
}

static void  PCI_OUTW(unsigned long ioaddr,unsigned short data)
{
        outl(ioaddr,PCI_PORT_ADDRESS);
        outw(data,PCI_PORT_DATA);
}

static unsigned char  PCI_INB(unsigned long ioaddr)
{
        outl(ioaddr,PCI_PORT_ADDRESS);
        return inb(PCI_PORT_DATA);
}

static void  PCI_OUTB(unsigned long ioaddr,unsigned char data)
{
        outl(ioaddr,PCI_PORT_ADDRESS);
        outb(data,PCI_PORT_DATA);
}


///////////////////////////////////////////////////////////////////////////////
// refer to PCI spec for follwing define and function

#define CONFIG_ADDRESS_BUS(addr) 		(addr>>16)&0x000000ff
#define CONFIG_ADDRESS_DEVICENUM(addr)		(addr>>11)&0x0000001f
#define CONFIG_ADDRESS_FUNCNUM(addr)		(addr>>8)&0x00000007


// this function find the device according to the deviceID and vendorID
// it find the first device and return the address
// so if will not find the second device if the system have this kind of device more than one.
// and also if the device have more than one function 
// then this funciton just find the first funciton and return the address
static unsigned long PCI_ScanDevice(unsigned short ProductID,unsigned short ManufacturerID)
{
        unsigned long laddress;
        unsigned long ID;
	
	ULONG	bus;
	ULONG	devicenum;
	ULONG	funcnum;

        for (laddress = 0x80000000 ; laddress <= 0x81000000 ; laddress += 0x100)
	{
                ID = PCI_IND(laddress);
                if (((unsigned short)(ID) == ManufacturerID) && ((unsigned short)(ID >> 16) == ProductID))
		{
			bus = CONFIG_ADDRESS_BUS(laddress);
			devicenum = CONFIG_ADDRESS_DEVICENUM(laddress);
			funcnum = CONFIG_ADDRESS_FUNCNUM(laddress);
			
			PDEBUG("bus %d device number %d function number %d,%x:%x",bus,devicenum,funcnum,ManufacturerID,ProductID);
                        return laddress;
                }
        }

        return 0;
}

// this function find the device according to the deviceID vendorID and function number
static unsigned long PCI_ScanDeviceFunction(unsigned short ProductID,unsigned short ManufacturerID,unsigned char FunctionID)
{
	unsigned long laddress;
        unsigned long ID;
	ULONG	bus;
	ULONG	devicenum;
	ULONG	funcnum;

        for (laddress = 0x80000000 ; laddress <= 0x81000000 ; laddress += 0x100)
	{// first find the device;search based on the diffirent device number 
                ID = PCI_IND(laddress);
                if (((unsigned short)(ID) == ManufacturerID) && ((unsigned short)(ID >> 16) == ProductID))
		{
			if((unsigned char)(CONFIG_ADDRESS_FUNCNUM(laddress))==FunctionID)
			{
				bus = CONFIG_ADDRESS_BUS(laddress);
				devicenum = CONFIG_ADDRESS_DEVICENUM(laddress);
				funcnum = CONFIG_ADDRESS_FUNCNUM(laddress);
				PDEBUG("address %x bus %d device number %d function number %d,%x:%x",laddress,bus,devicenum,funcnum,ManufacturerID,ProductID);
				return laddress;	
			}
                }
        }
	return 0;
}
static int  __devinit mp1_probe();

static void __devexit mp1_remove();


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)
static void InterruptHandler(int irq, void *dev_id, struct pt_regs *regs)
{
#ifdef INCLUDE_O2SD
	if(o2sd_check_intr())
	{
//		PDEBUG("interrupt from o2sd");
		o2sd_interrupt();
		return;
	}
#endif
#ifdef INCLUDE_O2MEDIA
	if(o2media_check_intr())
	{
//		PDEBUG("interupt from o2media");
		o2media_interrupt();
		return;
	}
#endif
	PDEBUG("interrupt from system");
}
#else
static irqreturn_t InterruptHandler(int irq, void *dev_id, struct pt_regs *regs)
{
	irqreturn_t ret = IRQ_NONE;
#ifdef INCLUDE_O2SD
	if(o2sd_check_intr())
	{
		o2sd_interrupt();
		ret |= IRQ_HANDLED;
	}
#endif
#ifdef INCLUDE_O2MEDIA
	if(o2media_check_intr())
	{
		o2media_interrupt();
		ret |= IRQ_HANDLED;
	}
#endif
	return ret;
}
#endif

UCHAR o2base_irq = 0;
static int presidio_f0_init(void)
{
        struct pci_dev     *pdev_f0 = NULL;


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
        if ( !pci_present () )
	{
                printk ("presidio: PCI BIOS not present\n");
                return -1;
        }
#endif
	// find the presidio function 0 device
	pdev_f0 = pci_find_device (VENDOR_O2MICRO,DEVICE_PRESIDIO_F0 , pdev_f0);
	if(pdev_f0!=NULL)
	{
         	pci_enable_device(pdev_f0);
//		PDEBUG("presidio: find o2micro presidio function 0 device");
	}
	else
	{
		PDEBUG("presidio: can not find o2micro presidio function 0 device");
		return -1;
	}
	// write the ODR configure register; detailed info please refer to spec.
	// now the work is done by BIOS
	/*
	pci_write_config_byte(pdev_f0,0xE0,0x04);		
	pci_write_config_byte(pdev_f0,0xE4,0x99);
	pci_write_config_byte(pdev_f0,0xE6,0x09);
	*/
	if(pdev_f0->irq>0 && (pdev_f0->irq!=0xFF))
	{
		o2base_irq = pdev_f0->irq;
//		PDEBUG("device function 0 irq = %x",o2base_irq);
	}
	else
	{
		pci_write_config_byte(pdev_f0,0x3C,11);
		o2base_irq = 11;
//		PDEBUG("manually set device funtion 0 irq = 11");
	}	
	return 0;
}

struct pci_dev     *pdev_f2 = NULL;
struct pci_dev     *pdev_f3 = NULL;
ULONG o2sd_base = 0;
UCHAR o2sd_irq = 0;
int o2sd_device_ID=0;
static int  presidio_f2_init(void)
{
	ULONG bus_address;
	ULONG bus_offset;
	pdev_f2 = pci_find_device (VENDOR_O2MICRO,DEVICE_PRESIDIO_F2 , pdev_f2);
	if(pdev_f2!=NULL)
	{
        	pci_enable_device(pdev_f2);
//		PDEBUG("presidio: find o2micro presidio function 2 device");
	}
	else
	{
		PDEBUG("presidio: can not find o2micro presidio function 2 device");
		return 0;
	}
//	pci_write_config_dword(pdev_f2,0x10,0xf1100000);
	bus_address = pci_resource_start(pdev_f2,0);
	// we must request mem regions first before use
	if(pci_request_regions(pdev_f2,"O2SD")<0)
	{
		printk("o2sd request_mem_region failed");
		return 0;
	}
//	bus_address 	&= 	PCI_BASE_ADDRESS_MEM_MASK;  // #define  PCI_BASE_ADDRESS_MEM_MASK (~0x0fUL) in <linux/pci.h>
//	bus_address 		= 	bus_address & ~(PAGE_SIZE - 1);  	// #define PAGE_SIZE 0x400  in <linux/a.out.h>
//	bus_offset 		= 	bus_address &   (PAGE_SIZE - 1);
//	PDEBUG( "SDIO_BASE_Before_IO_Mapt=%X", bus_address);
	o2sd_base = (ULONG)(ULONG*)ioremap_nocache(bus_address, pci_resource_len(pdev_f2,0));
//	o2sd_base = (ULONG)(ULONG*)ioremap_nocache(0xf1100000, pci_resource_len(pdev_f2,0));
	if(!o2sd_base)
	{
		PDEBUG("ioremap failed");
		goto FAILED01;
	}
	
	if(pdev_f2->irq>0 && (pdev_f2->irq!=0xFF))
	{
		o2sd_irq = pdev_f2->irq;
//		PDEBUG("o2sd_irq = %x",o2sd_irq);
	}
	else
	{
		o2sd_irq = o2base_irq;
		pci_write_config_byte(pdev_f2,0x3C,o2base_irq);
		pdev_f2->irq = o2sd_irq;
        	pci_enable_device(pdev_f2);
//		PDEBUG("manually set function 2 irq = %x",o2base_irq);
	}
	if (request_irq (o2sd_irq, InterruptHandler, SA_SHIRQ,   (const char *) DRIVER_NAME,  (void *)(&o2sd_device_ID)))
	{
		PDEBUG("request irq %x for o2sd failed",o2sd_irq);
		goto FAILED0;
	}
	o2sd_SCIInitialize(o2sd_base,0);
	return 0;
FAILED0:
	iounmap((void*)o2sd_base);
FAILED01:
	pci_release_regions(pdev_f2);
	return -1;

}


ULONG o2media_base = 0;
UCHAR o2media_irq = 0;
int o2media_device_ID=0;
static int  presidio_f3_init(void)
{
	ULONG bus_address;
	ULONG bus_offset;
	pdev_f3 = pci_find_device (VENDOR_O2MICRO,DEVICE_PRESIDIO_F3 , pdev_f3);
	if(pdev_f3!=NULL)
	{
        	pci_enable_device(pdev_f3);
//		PDEBUG("presidio: find o2micro presidio function 3 device");
	}
	else
	{
		PDEBUG("presidio: can not find o2micro presidio function 3 device");
		return 0;
	}
	bus_address = pci_resource_start(pdev_f3,0);
	if(pci_request_regions(pdev_f3,"O2MEDIA")<0)
	{
		printk("o2media request_mem_region failed");
		return 0;
	}
//	bus_address 	&= 	PCI_BASE_ADDRESS_MEM_MASK;  // #define  PCI_BASE_ADDRESS_MEM_MASK (~0x0fUL) in <linux/pci.h>
//	bus_address 		= 	bus_address & ~(PAGE_SIZE - 1);  	// #define PAGE_SIZE 0x400  in <linux/a.out.h>
//	bus_offset 		= 	bus_address &   (PAGE_SIZE - 1);
//	PDEBUG( "O2media_BASE_Before_IO_Mapt=%X", bus_address);
	o2media_base = (ULONG)(ULONG*)ioremap_nocache(bus_address, pci_resource_len(pdev_f3,0));
	if(!o2media_base)
	{
		PDEBUG("ioremap failed");
		goto FAILED11;
		
	}
	if(pdev_f3->irq>0 && (pdev_f3->irq!=0xFF))
	{
		o2media_irq = pdev_f3->irq;
//		PDEBUG("o2media_irq = %x",o2media_irq);
	}
	else
	{
		o2media_irq = o2base_irq;
		pci_write_config_byte(pdev_f3,0x3C,o2base_irq);
		pdev_f3->irq = o2media_irq;
        	pci_enable_device(pdev_f3);
//		PDEBUG("manually set function 3 irq = %x",o2base_irq);
	}
#ifndef INCLUDE_O2SD
// we use this macro to avoid register this handler twice.
	if (request_irq (o2media_irq, InterruptHandler, SA_SHIRQ,   (const char *) DRIVER_NAME,  (void *)(&o2media_device_ID)))
	{
		PDEBUG("request irq %x for o2media failed",o2media_irq);
		goto FAILED1;
	}
#endif
	o2media_SCIInitialize(o2media_base,0);
	return 0;
FAILED1:
	iounmap((void*)o2media_base);
FAILED11:
	pci_release_regions(pdev_f3);
	return -1;
}

static void ReleaseResourceF2(void)
{
	if(o2sd_base)
	{
	free_irq(o2sd_irq,(void*)&o2sd_device_ID);
	iounmap((void*)o2sd_base);
	o2sd_base = 0;
	}
	pci_release_regions(pdev_f2);
}
static void ReleaseResource(void)
{
#ifdef INCLUDE_O2SD
	ReleaseResourceF2();	
#endif
#ifndef INCLUDE_O2SD
	if(o2media_base)
	free_irq(o2media_irq,(void*)&o2media_device_ID);
#endif
#ifdef INCLUDE_O2MEDIA
	if(o2media_base)
	{
	iounmap((void*)o2media_base);
	o2media_base = 0;
	}
	pci_release_regions(pdev_f3);
#endif
}

#define pm_message_t unsigned long 
#define pci_power_t unsigned long


static int __devinit mp1_probe(struct device *pdev)
{
//	if(presidio_f0_init())
//		return 0;
#ifdef INCLUDE_O2SD
	presidio_f2_init();
#endif
#ifdef INCLUDE_O2MEDIA
	if(presidio_f3_init())
	{
#ifdef INCLUDE_O2SD
		ReleaseResourceF2();
#endif
		return 0;
	}
#endif
	return 0;
}
static void __devexit mp1_remove(struct device* pdev)
{
#ifdef INCLUDE_O2SD
	if(o2sd_base)
	o2sd_preUnInit();
#endif
#ifdef INCLUDE_O2MEDIA
	if(o2media_base)
	o2media_preUnInit();	
#endif
	ReleaseResource();
	return;
}
#ifdef CONFIG_PM
static int mp1_suspend(struct device *pdev, pm_message_t state,u32 level)
{
	if(level == SUSPEND_DISABLE)
	{
#ifdef INCLUDE_O2SD
		o2sd_suspend();
#endif
#ifdef INCLUDE_O2MEDIA
		o2media_suspend();
#endif
	}
	return 0;
}
static int mp1_resume(struct device *pdev,u32 level)
{
	if(level == RESUME_POWER_ON)
	{
#ifdef INCLUDE_O2SD
		o2sd_resume();
#endif
#ifdef INCLUDE_O2MEDIA
		o2media_resume();
#endif
	}
	return 0;
}
#endif
static struct platform_device *mp1_platform_device;
static struct device_driver mp1_driver = {
       .name           = "mp1",
       .bus          = &platform_bus_type,
	.probe		= mp1_probe,
	.remove		= mp1_remove,
#ifdef CONFIG_PM
       .suspend        = mp1_suspend,
       .resume         = mp1_resume,
#endif

};

static int __init presidio_init(void)
{
	int i = 0;
	mp1_platform_device = platform_device_register_simple("mp1",-1,NULL,0);
	if(IS_ERR(mp1_platform_device))
		return PTR_ERR(mp1_platform_device);
	i = driver_register(&mp1_driver);
	if(i<0)
	{
		platform_device_unregister(mp1_platform_device);
		printk("driver_register failed\n");
		return i;
	}
//	if(mp1_probe())
//	{
//		platform_device_unregister(mp1_platform_device);
//		driver_unregister(&mp1_driver);
//	}
	return i;

}

static void __exit presidio_exit(void)
{
	driver_unregister(&mp1_driver);
	platform_device_unregister(mp1_platform_device);
//	mp1_remove();
}
/*
// function to keep insmod happy
void __pure_virtual(void)
{
}
*/
// function to keep insmod happy
void __cxa_pure_virtual(void)
{
}

/*
int __cxa_guard_acquire (char *g) 
{
   return !*(char *)(g);
}
 
void __cxa_guard_release (char *g)
{
    *(char *)g = 1;
}


void __cxa_guard_acquire(void)
{
}
void __cxa_guard_release(void)
{
}
*/
// this function is needed by gcc high version for c++
  __extension__ typedef int __guard __attribute__((mode (__DI__)));
int __cxa_guard_acquire (__guard *g) 
  {
    return !*(char *)(g);
  }


  void __cxa_guard_release (__guard *g)
  {
    *(char *)g = 1;
  }


module_init(presidio_init);
module_exit(presidio_exit);
MODULE_DESCRIPTION("O2Micro presidio device driver");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.3");
