/*
 * OZ711MS1 OZ711MP1 linux driver
 * O2Micro Inc.
 * toby.yang@o2micro.com
 * 2005/4/17
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*
 *  A OZSCR PCMCIA client driver
 * Revisions 1.2.2
 * O2Micro Inc.
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "base.h"
#include "basic.h"
#include "o2sd_main.h"
#include "scsi_o2sd.h"
#include "ellen/sdioif.h"
#define winif_PDEBUG(fmt,args...) printk("presidio: " fmt "\n" , ## args)
//#define winif_PDEBUG
static sdio_host_desc* g_psdhostdesc = NULL;
static sdio_card_desc* g_psdcard = NULL;

static BOOLEAN o2sd_chkMediaExist(sdio_card_desc* card)
{
	return sdio_if_chkMediaExist(card);
}

static BOOLEAN o2sd_chkWriteProtect(sdio_card_desc* card)
{
	return sdio_if_chkWriteProtect(card);
}

static  BOOLEAN o2sd_host_stop(sdio_host_desc* pHostDevice)
{
	sdio_if_cleanup(pHostDevice);
	return TRUE;
}


// o2sd_thread is created for scsi register and unregister when card insert 
// or remove.
// we can not do this in interrupt handler for the scsi mid level driver will call "kernel_thread" . and "kernel_thread" will call "schedule()" which is not allowed in interrupt handler. 
static void o2sd_thread(PVOID context)
{
	sdio_card_desc *temp = (sdio_card_desc*)context;
//	winif_PDEBUG("context=%x",context);
	do
	{
		o2sd_wait_event_interruptible();
		if(o2sd_check_intrflags()==TRUE)
		{
			o2sd_clear_intrflags();
			if(o2sd_chkMediaExist(temp))
			{// card insert
				winif_PDEBUG("card insert ");
				if(temp->o2sd_scsi_has_registered==0)
				{
					o2sd_scsi_register((unsigned long)context);
					temp->o2sd_scsi_has_registered = 1;
//					winif_PDEBUG("scsi register for card insert");
				}
				else
					winif_PDEBUG("error:card insert but scsi has registered before");
			}
			else
			{// card remove
				winif_PDEBUG("card remove");
				if(temp->o2sd_scsi_has_registered==1)
				{
					o2sd_scsi_unregister();
					temp->o2sd_scsi_has_registered = 0;
//					winif_PDEBUG("scsi unregister for card remove");
			// if some error happen,the controller will stay in the wrong status.
			// so we reinit the controller when card remove for maybe there are some error when unexpected card removing.
					sdio_if_init(temp->membase,temp->pHostDevice);
				}
				else
					winif_PDEBUG("error:card remove but scsi has not been registered before");
			}
		}
		else
		{
			winif_PDEBUG("thread receive terminate signal");
			break;
		}
		
	}while(!winif_signal_pending_current());
	// for thread receiving terminate signal, we must unregistered scsi.
	if(temp->o2sd_scsi_has_registered)
	{
		o2sd_scsi_unregister();
		temp->o2sd_scsi_has_registered = 0;
		winif_PDEBUG("scsi unregister for rmmod call");
	}
	o2sd_complete_and_exit();
}

static void o2sd_preInit(sdio_card_desc* ptemp_devext)
{
	int pid; 
	pid = winif_kernel_thread((void*)o2sd_thread,(PVOID)ptemp_devext);
	if (pid>=0)
	{
		o2sd_pri_data.ko2sd_pid = pid;
		return;
	}
	else
	{
		winif_PDEBUG("failed to create o2sd thread for scsi (un)register");
		return;
	}
}

void o2sd_preUnInit(void)
{
	winif_kill_proc(o2sd_pri_data.ko2sd_pid,SIGKILL);
	o2sd_wait_for_completion();
	o2sd_host_stop(g_psdhostdesc);
#ifdef SCSI_THREAD
	o2sd_scsi_preuninit();
#endif
//	winif_PDEBUG("after scsi_preuninit");
	kfree(g_psdcard);
	kfree(g_psdhostdesc);
}



BOOLEAN o2sd_bus_reset(void)
{
	return SUCCESS;
}


BOOLEAN o2sd_device_reset(void)
{
//	return SUCCESS;
	return sdio_if_cardreset(g_psdcard);
}

BOOLEAN o2sd_cmd_abort(void)
{
	return SUCCESS;
}
BOOLEAN o2sd_check_intr(void)
{      
	return sdio_if_interrupt_claim( g_psdhostdesc);
}

BOOLEAN o2sd_interrupt(void)
{
	return sdio_if_interrupt_handler(g_psdhostdesc);
}

BOOLEAN o2sd_suspend()
{
	if(g_psdcard==NULL)
	{	
		printk("g_psdcard is NULL when suspend\n");
		return FALSE;
	}
	else
		return sdio_if_host_suspend(g_psdcard);
}
BOOLEAN o2sd_resume()
{
	if(g_psdcard==NULL)
	{
		printk("g_psdcard is NULL when resume\n");
		return FALSE;
	}
	else
	{
		sdio_if_host_resume(g_psdcard);
		return o2sd_SCIInitialize(g_psdcard->membase,0);
	}
}

BOOLEAN o2sd_SCIInitialize(ULONG m_o2sd_base,ULONG resume)
{
	if(g_psdhostdesc==NULL)
	{
		g_psdhostdesc = kmalloc(sizeof(sdio_host_desc),GFP_KERNEL);
//		winif_PDEBUG("g_psdhost malloc at %x",g_psdhostdesc);
	}
	sdio_if_init(m_o2sd_base,g_psdhostdesc);
	if(g_psdcard==NULL)
	{
		g_psdcard = kmalloc(sizeof(sdio_card_desc),GFP_KERNEL);
//		winif_PDEBUG("g_psdcard malloc at %x",g_psdcard);
	}
#ifdef SCSI_THREAD
	if(resume==0)
	o2sd_scsi_preinit();
#endif
	g_psdcard->pHostDevice = g_psdhostdesc;
	g_psdcard->o2sd_scsi_has_registered = 0;
        sdio_if_getcardinfo ( g_psdcard );

	if(resume==0)
	o2sd_preInit(g_psdcard);
	if(sdio_if_chkMediaExistTrue(g_psdcard)) {
		o2sd_set_cardexist();
		o2sd_set_intrflags();
	}
	return TRUE;
}


BOOLEAN o2sd_host_reset(unsigned long devext_addr)
{
//	return SUCCESS;
	return sdio_if_hostreset(((sdio_card_desc*)devext_addr)->pHostDevice);
}


int o2sd_inquiry(INQUIRY_DATA* inquiry,unsigned long devext_addr)
{			
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
	if (!o2sd_chkMediaExist(ptemp))
	{	
		inquiry->DeviceType = 0x7f;//TYPE_NO_LUN;
		return 1;
	}
	else
	{
		inquiry->DeviceType = 0x00;//TYPE_DISK;
		inquiry->DeviceTypeQualifier = 0x80;// removable 
		inquiry->Version = 0;
		inquiry->ResponseDataFormat = 2;
		inquiry->AdditionalLength = 31;
		inquiry->Flags[0] = 0x1;
		inquiry->Flags[1] = 0x32;
		winif_strncpy((char*)(inquiry->VendorId), "O2Micro ", 8);
		winif_strncpy((char*)(inquiry->ProductId), "O2SD            ", 16);
		winif_strncpy((char*)(inquiry->ProductRevisionLevel), "0001", 4);
		//	SCpnt->result = DID_OK << 16;
		return 0;
	}
}

int o2sd_testunitready(unsigned long devext_addr)
{
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
	if (!o2sd_chkMediaExist(ptemp))
	{
		return 2;	
	}
	else
		return 0;
}

int o2sd_requestsense(SENSE_DATA *senseBuffer,unsigned long devext_addr)
{
	int ret = 0;
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
//	printk("in requestsense\n");
	ret = o2sd_chkMediaExist(ptemp);
			if (ret)
			{
#ifdef ENABLE_WRITE_PROTECT
				if(o2sd_chkWriteProtect(ptemp))
				{
//					printk("write protect in requestsense\n");
					senseBuffer->ResponseCode = 0xF0;
					senseBuffer->AdditionalSenseLength = 0xb;
					senseBuffer->SenseKey =  0x07;
					senseBuffer->AdditionalSenseCode = 0x27;
					senseBuffer->AdditionalSenseCodeQual = 0;
				}
				else
#endif
				{
//				winif_PDEBUG(" - senseType == MEDIA_IN_READER. ");
					senseBuffer->ResponseCode = 0xF0;
					senseBuffer->AdditionalSenseLength = 0xb;
					senseBuffer->SenseKey =  0x06;//UNIT_ATTENTION;
					senseBuffer->AdditionalSenseCode = 0;
					senseBuffer->AdditionalSenseCodeQual = 0;
				}
			}
			else 
			{
//				winif_PDEBUG(" - senseType == NO_MEDIA_IN_READER. ");
				senseBuffer->ResponseCode = 0xF0;
				senseBuffer->AdditionalSenseLength = 0xb;
				senseBuffer->SenseKey =  0x02;//NOT_READY;
				senseBuffer->AdditionalSenseCode = 0x3a;
				senseBuffer->AdditionalSenseCodeQual = 0;
			}
			return 3;
}

int o2sd_modesense(MODE_PAGE_DATA* mdata,BOOLEAN mdata_3f,unsigned long devext_addr)
{
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
//	printk("in modesense mdata_3f=%d\n",mdata_3f);
			if (mdata_3f == TRUE)
			{
				mdata->hdr.DataLength = sizeof(MODE_PAGE8) + 3;
				mdata->hdr.MediumType = 0x00;
			// set for write protect
#ifdef ENABLE_WRITE_PROTECT 
				if(o2sd_chkWriteProtect(ptemp))
				{
//					printk("write protect in modesense\n");
					mdata->hdr.Reserved = 0x80;
				}
				else
					mdata->hdr.Reserved = 0;
#endif			
				mdata->hdr.BlockDescLength = 0;
				mdata->pdata.pg8.PageCode = 8;
				mdata->pdata.pg8.PageLength = 0x12;
				return 0;
			}
			else 
			{
				return 2;
			}
}
int  o2sd_allowmediumremoval(ULONG devext_addr)
{
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
	if (!o2sd_chkMediaExist(ptemp))
	{
		return 2;	
	}
	else
		return 0;
}

#ifndef BIG_ENDIAN_ULONG
#define BIG_ENDIAN_ULONG(x)		((x&0xff000000) >> 24) | \
								((x&0x00ff0000) >> 8 ) | \
								((x&0x0000ff00) << 8 ) | \
								((x&0x000000ff) << 24)
#endif ///BIG_ENDIAN_ULONG
int o2sd_readcapacity(READ_CAPACITY_DATA* capacityData,ULONG devext_addr)
{
	sdio_card_desc* ptemp = (sdio_card_desc*)devext_addr;
	int	ret=0;
	USHORT		Cyl; 
	UCHAR		Head; 
	UCHAR		Sect;
	ULONG		Capacity;
	if (!o2sd_chkMediaExist(ptemp))
	{
		return 1;	
	}
	if(!sdio_if_getcardinfo(ptemp))
	{
		capacityData->LogicalBlockAddress =BIG_ENDIAN_ULONG(ptemp->LBACount - 1);
		capacityData->BytesPerBlock = BIG_ENDIAN_ULONG(ptemp->BlockSize);
		ret = 0;
	}
	else
	{
		capacityData->LogicalBlockAddress = 0;
		capacityData->BytesPerBlock = 0;
		ret = 1;
	}
	return ret;
	
}

int	o2sd_readData(PUCHAR currentBuf,ULONG startBlock,ULONG blockCount,ULONG devext_addr)
{
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
	int	ret = 0;
	if (!o2sd_chkMediaExist(ptemp))
	{
		return 1;	
	}
	if (blockCount==0)
		return 0;
	ret = sdio_if_readsector( ptemp,currentBuf,startBlock, blockCount);
	if(ret!=blockCount*(ptemp->BlockSize))
		ret = 1;
	else
		ret = 0; 
	return ret;
}


int o2sd_writeData(PUCHAR currentBuf,ULONG startBlock,ULONG blockCount,ULONG devext_addr)
{
	sdio_card_desc *ptemp = (sdio_card_desc*)devext_addr;
	int	ret = 0;
	if (!o2sd_chkMediaExist(ptemp))
	{
		return 1;	
	}
#ifdef ENABLE_WRITE_PROTECT
	if (o2sd_chkWriteProtect(ptemp))
	{
		printk("check condition for write protect\n"); 
		return 2;
	}
#endif
	if (blockCount==0)
		return 0;
	ret = sdio_if_writesector( ptemp,currentBuf,startBlock, blockCount);
	if(ret!=blockCount*(ptemp->BlockSize))
		ret = 1;
	else
		ret = 0; 
	return ret;
}

