#include "basic.h"
#include "base.h"
#include "scsi_o2media.h"
#include "scsi_base.h"
#include "winif2.h"
static int o2media_cmd_handler(struct scsi_cmnd *SCpnt , void (*done)(struct scsi_cmnd *));
static unsigned long my_o2media_devext_addr;


struct t_o2media_pri_data o2media_pri_data;

/***********************************************************************
 * Host functions 
 ***********************************************************************/
static const char* host_info(struct Scsi_Host *host)
{
	return "SCSI emulation for o2media devices";
}



#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
/* detect a virtual adapter (always works) */
static int detect(struct SHT *sht)
{
	struct Scsi_Host *host;
	host = scsi_register(sht, sizeof(my_o2media_devext_addr));
	if(host)
	{
		host->hostdata[0] = my_o2media_devext_addr;
		host->irq = o2media_irq;
		host->base = o2media_base;
		host->max_id = 1;
		host->max_lun = 1;
		return 1;
	}
	else
		return 0;
}
#endif

static int release(struct Scsi_Host *psh)
{
	/* we always have a successful release */
	// here psh->hostt must == &o2media_host_template;
	if(psh->hostt != o2media_pri_data.po2media_myht)
		return -1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	// we clear the o2media_host_template->next.
	// for the scsi_register_module for next time will check this flag.
	o2media_pri_data.po2media_myht->next = 0;
#else
#endif
	return 0;
}

/* run command */
/* run command */
/*
static void cmd_finished(Scsi_Cmnd *SCpnt)
{
     SCpnt->SCp.Status++;
}
static int command( Scsi_Cmnd *srb )
{
	PDEBUG("in command");
	queuecommand(srb,cmd_finished);
	while(!srb->SCp.Status)
		PDEBUG("In Command Handler....");
	return srb->result;
}
*/

/***********************************************************************
 * Error handling functions
 ***********************************************************************/

/* Command abort */
static int command_abort( struct scsi_cmnd *srb )
{
	setscsiresult(srb,0);
	srb->scsi_done(srb);
	return SUCCESS;
}

/* This invokes the transport reset mechanism to reset the state of the
 * device */
static int device_reset( struct scsi_cmnd *srb )
{
	setscsiresult(srb,0);
	srb->scsi_done(srb);
	return SUCCESS;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	return o2media_device_reset(srb->host->hostdata[0]);
#else
	return o2media_device_reset((unsigned long)(srb->device->host->hostdata[0]));
#endif
}

/* This resets the device port, and simulates the device
 * disconnect/reconnect for all drivers which have claimed other
 * interfaces. */
static int bus_reset( struct scsi_cmnd *srb )
{
	return TRUE;
}

/* FIXME: This doesn't do anything right now */
static int host_reset( struct scsi_cmnd *srb )
{
	return FALSE;
}

static void o2media_cmd_control_thread(void* pri_data)
{
	unsigned long flags;
	struct t_o2media_pri_data* x = (struct t_o2media_pri_data*)(pri_data);
	do
	{
		wait_event_interruptible(x->o2media_cmd_control_event,x->o2media_cmd_comming);
		if(x->o2media_cmd_comming == 1)
		{
			if(x->cur_srb != NULL)
			{
			o2media_cmd_handler(x->cur_srb,x->cur_srb->scsi_done);
			spin_lock_irqsave(&(x->cmd_lock),flags);
			x->o2media_cmd_comming = 0;
			spin_unlock_irqrestore(&(x->cmd_lock),flags);
			spin_lock_irqsave(&(x->srb_lock),flags);
			x->cur_srb = NULL;		
			spin_unlock_irqrestore(&(x->srb_lock),flags);
			}
		}
		else
		{
			PDEBUG("o2media cmd control thread receive terminate signal");
			break;
		}
			
	}while(!winif_signal_pending_current());
	// for thread receiving terminate signal
	complete_and_exit(&(x->notify),0);
}


void o2media_scsi_preinit()
{
	
	// the speed is slow, we try to use thread to handler scsi command and return ok when receiving command. 
	// we also must use thread to handle read and write for our read and write may be sleep for hardware going but the read and write will be called by rw_intr--->.....which is in interrupt context.
	int pid;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	o2media_pri_data.po2media_myht=NULL;
#else
	o2media_pri_data.o2media_sh=NULL;
	o2media_pri_data.po2media_myht=NULL;
#endif
	o2media_pri_data.ko2media_cmd_control_pid = 0;
	o2media_pri_data.o2media_cmd_comming = 0;
	o2media_pri_data.cmd_lock = SPIN_LOCK_UNLOCKED;
	o2media_pri_data.srb_lock = SPIN_LOCK_UNLOCKED;
	o2media_pri_data.buffer = NULL;
	o2media_pri_data.cur_srb = NULL;
	o2media_pri_data.buffer_size = 0;
	init_waitqueue_head(&(o2media_pri_data.o2media_cmd_control_event));
	init_completion(&(o2media_pri_data.notify));
#ifndef INSREM_CMD_THREAD	
	pid = winif_kernel_thread((void*)o2media_cmd_control_thread,(PVOID)&o2media_pri_data);
	if(pid>=0)
		o2media_pri_data.ko2media_cmd_control_pid = pid;
	else
	{
		PDEBUG("create o2media cmd control thread failed");
		return;
	}
#endif	
}

static void free_inside_buffer()
{
	vfree(o2media_pri_data.buffer);
	o2media_pri_data.buffer = NULL;
	o2media_pri_data.buffer_size = 0;
}
void o2media_scsi_preuninit()
{
#ifndef INSREM_CMD_THREAD	
	if(o2media_pri_data.ko2media_cmd_control_pid)	
		winif_kill_proc(o2media_pri_data.ko2media_cmd_control_pid,SIGKILL);
	o2media_pri_data.ko2media_cmd_control_pid=0;
	wait_for_completion(&(o2media_pri_data.notify));
#endif
	/*
	if(o2media_pri_data.po2media_myht);
	{
		kfree(o2media_pri_data.po2media_myht);
		o2media_pri_data.po2media_myht = NULL;
	}
	*/
	free_inside_buffer();
	
}
static void malloc_inside_buffer(unsigned long size,unsigned long blocksize)
{
	if(!o2media_pri_data.buffer)
	{
		o2media_pri_data.buffer = vmalloc(size);
		if(!o2media_pri_data.buffer)
			PDEBUG("vmalloc for inside buffer %d failed",size);		
		o2media_pri_data.buffer_size = size;
	}
	if(o2media_pri_data.buffer_size < size)
	{
		vfree(o2media_pri_data.buffer);
		o2media_pri_data.buffer = vmalloc(size);
		if(!o2media_pri_data.buffer)
			PDEBUG("vmalloc for inside buffer %d failed",size);		
		o2media_pri_data.buffer_size = size;
	}
		
}

void o2media_scsi_register(unsigned long devext_addr)
{
	int ret;
	
	int pid;
#ifdef INSREM_CMD_THREAD
	pid = winif_kernel_thread((void*)o2media_cmd_control_thread,(PVOID)&o2media_pri_data);
	if(pid>=0)
		o2media_pri_data.ko2media_cmd_control_pid = pid;
	else
	{
		PDEBUG("create o2media cmd control thread failed");
		return;
	}
#endif	
	my_o2media_devext_addr = devext_addr;
//	PDEBUG("in scsi register");
//	PDEBUG("devext_addr = %x",devext_addr);
#ifdef SCSI_THREAD
//	o2media_scsi_preinit();
#endif
		
	if(o2media_pri_data.po2media_myht==NULL)
	{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		o2media_pri_data.po2media_myht = (Scsi_Host_Template*)kmalloc(sizeof(Scsi_Host_Template),GFP_KERNEL);
		memset(o2media_pri_data.po2media_myht,0,sizeof(Scsi_Host_Template));
#else
		o2media_pri_data.po2media_myht = (struct scsi_host_template*)kmalloc(sizeof(struct scsi_host_template),GFP_KERNEL);
		memset(o2media_pri_data.po2media_myht,0,sizeof(struct scsi_host_template));
#endif
	}
	memcpy(o2media_pri_data.po2media_myht, &o2media_host_template,sizeof(o2media_host_template));
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	ret = scsi_register_module(MODULE_SCSI_HA,o2media_pri_data.po2media_myht);
	if(ret)
		PDEBUG("scsi_register_module failed");
#else	
	o2media_pri_data.o2media_sh = scsi_host_alloc(o2media_pri_data.po2media_myht,sizeof(my_o2media_devext_addr));
	if(!o2media_pri_data.o2media_sh)
	{
		PDEBUG("Unable to register controller with SCSI subsystem");
		return -1;
	}
	o2media_pri_data.o2media_sh->hostdata[0] = my_o2media_devext_addr;
	o2media_pri_data.o2media_sh->irq = o2media_irq;
	o2media_pri_data.o2media_sh->base = o2media_base;
	o2media_pri_data.o2media_sh->max_id = 1;
	o2media_pri_data.o2media_sh->max_lun = 1;

	scsi_add_host(o2media_pri_data.o2media_sh,NULL);
	scsi_scan_host(o2media_pri_data.o2media_sh);
//	PDEBUG("end of scsi register");
#endif
}

void o2media_scsi_unregister(void)
{
	int ret;
#ifdef INSREM_CMD_THREAD
	winif_kill_proc(o2media_pri_data.ko2media_cmd_control_pid,SIGKILL);
	o2media_pri_data.ko2media_cmd_control_pid=0;
	wait_for_completion(&(o2media_pri_data.notify));
#endif
#ifdef SCSI_THREAD
//	o2media_scsi_preuninit();
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
			ret = scsi_unregister_module(MODULE_SCSI_HA,o2media_pri_data.po2media_myht);
			if(ret)
				PDEBUG("scsi_unregister_module failed");
#else
			scsi_remove_host(o2media_pri_data.o2media_sh);
#endif
	// we can not free po2media_myht for system will use when user umount
	// we free it when driver unload.
			scsi_host_put(o2media_pri_data.o2media_sh);
//			kfree(o2media_pri_data.po2media_myht);
//			o2media_pri_data.po2media_myht = NULL;
//			free_inside_buffer();
	
			o2media_del_obj();
}

void o2media_scsicallback(void)
{
	/*
	if(o2media_current_srb && o2media_current_srb->scsi_done)
	{
		o2media_current_srb->scsi_done(o2media_current_srb);	
		o2media_current_srb = NULL;
	}
	*/
}
static int rtest(unsigned long start,unsigned long count,unsigned long devext)
{
	unsigned char* buf;
	unsigned long i;
	unsigned long j;
	buf = (unsigned char*)kmalloc(count*512,GFP_KERNEL);

	o2media_readData(buf,start,count,devext);
	for(j=0;j<count;j++)
	{
		dumpbuffer(buf+j*512,512);
	}		
	kfree(buf);
	return 0;
}
static int rwtest(unsigned long start,unsigned long count,unsigned long devext)
{
	unsigned char* ptr;
	unsigned char* buf;
	unsigned long i;
	ptr = (unsigned char*)kmalloc(count*512,GFP_KERNEL);
	buf = (unsigned char*)kmalloc(count*512,GFP_KERNEL);
	for(i = 0;i<count*512;i++)
	{
		*(ptr+i) =0x35; 
	}

	o2media_writeData(ptr,start,count,devext);
	o2media_readData(buf,start,count,devext);
	for(i = 0;i<count*512;i++)
	{
		if(*(buf+i)!=0x35)
		{
			PDEBUG("rwtest failed at %d=%x",i,*(buf+i));
			goto failed_rwtest;
		}
	}
	kfree(ptr);
	kfree(buf);
	return 0;
failed_rwtest:
	kfree(ptr);
	kfree(buf);
	return -1;

}

static int o2media_cmd_handler(struct scsi_cmnd *SCpnt , void (*done)(struct scsi_cmnd *))
{
	int ret;
	unsigned char* ptr;
	struct scatterlist* sg;
	unsigned long total = 0;
			int i;
	ULONG block,blocks;
	INQUIRY_DATA inquiry;
	SENSE_DATA  senseBuffer;
	MODE_PAGE_DATA mdata;
	READ_CAPACITY_DATA capacityData;
	BOOLEAN mdata_3f;
	UCHAR	*cdb = (UCHAR *)SCpnt->cmnd;// Pointer to SCSI CDB
	SCpnt->scsi_done = done;
	if (done)
	{
	}	
	else
	{
		PDEBUG("o2media_cmd_handler: %02X: done can't be NULL\n", *cdb);
		return 0;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	unsigned long devext_addr = (SCpnt->host->hostdata[0]);
#else
	unsigned long devext_addr = (SCpnt->device->host->hostdata[0]);
#endif	
		
//	PDEBUG("devext_addr before call cmd handler %x",devext_addr);
	switch ( *cdb )
	{
		case INQUIRY:// inquiry CDB
			if(!(SCpnt->request_buffer))
			{
				PDEBUG("request_buffer is NULL when inquiry cmd\n");
				setscsiresult(SCpnt,1);
				done (SCpnt);
				return 0;
			}
				
			memset(&inquiry, 0, sizeof(INQUIRY_DATA));
			ret = o2media_inquiry(&inquiry,devext_addr);
			if(SCpnt->use_sg)
			{
				ptr = &inquiry;
				sg = (struct scatterlist *) SCpnt->request_buffer;
				mp1_xfer_buf(SCpnt->use_sg,sg,ptr,1,0);
			}
			else
			memcpy(SCpnt->request_buffer, &inquiry, SCpnt->request_bufflen);
			setscsiresult(SCpnt,ret);
			done (SCpnt);
		return 0;
		case TEST_UNIT_READY:// test unit ready CDB
			ret = o2media_testunitready(devext_addr);
			setscsiresult(SCpnt,ret);
			done (SCpnt);
		return 0;
		case REQUEST_SENSE:
			if(!(SCpnt->request_buffer))
			{
				PDEBUG("request_buffer is NULL when request sense cmd\n");
				setscsiresult(SCpnt,1);
				done (SCpnt);
				return 0;
			}
			memset(&senseBuffer,0,sizeof(SENSE_DATA));
			ret = o2media_requestsense(&senseBuffer,devext_addr);
			if(SCpnt->use_sg)
			{
				ptr = &senseBuffer;
				sg = (struct scatterlist *) SCpnt->request_buffer;
				mp1_xfer_buf(SCpnt->use_sg,sg,ptr,1,0);
			}
			else
			memcpy(SCpnt->request_buffer, &senseBuffer, SCpnt->request_bufflen);
			setscsiresult(SCpnt,ret);
			done (SCpnt);
		return 0;
		case MODE_SENSE:
			if(!(SCpnt->request_buffer))
			{
				PDEBUG("request_buffer is NULL when mode sense cmd\n");
				setscsiresult(SCpnt,1);
				done (SCpnt);
				return 0;
			}
			if ((SCpnt->cmnd[2] & 0x3F) == 0x3F)
				mdata_3f = TRUE;
			else
				mdata_3f = FALSE;
			memset(&mdata, 0, sizeof(MODE_PAGE_DATA));
			ret = o2media_modesense(&mdata,mdata_3f,devext_addr);
			if(SCpnt->use_sg)
			{
				ptr = &mdata;
				sg = (struct scatterlist *) SCpnt->request_buffer;
				mp1_xfer_buf(SCpnt->use_sg,sg,ptr,1,0);
			}
			else
			memcpy(SCpnt->request_buffer, &mdata, SCpnt->request_bufflen);
			setscsiresult(SCpnt,ret);
			done (SCpnt);
		return 0;
		case READ_CAPACITY:
			if(!(SCpnt->request_buffer))
			{
				PDEBUG("request_buffer is NULL when readcapa cmd\n");
				setscsiresult(SCpnt,1);
				done (SCpnt);
				return 0;
			}
			memset(&capacityData,0,sizeof(READ_CAPACITY_DATA));
			ret = o2media_readcapacity(&capacityData,devext_addr);
			if(SCpnt->use_sg)
			{
				ptr = &capacityData;
				sg = (struct scatterlist *) SCpnt->request_buffer;
				mp1_xfer_buf(SCpnt->use_sg,sg,ptr,1,0);
			}
			else
			memcpy(SCpnt->request_buffer, &capacityData, SCpnt->request_bufflen);
			setscsiresult(SCpnt,ret);
			done(SCpnt);
	//		PDEBUG("readCapa OK");
	//		setcurrentsrb(SCpnt);
		return 0;
		case READ_10:
			if(!(SCpnt->request_buffer))
			{
				PDEBUG("request_buffer is NULL when read cmd\n");
				setscsiresult(SCpnt,1);
				done (SCpnt);
				return 0;
			}
			block = ((u32)(SCpnt->cmnd[2]) << 24) | ((u32)(SCpnt->cmnd[3]) << 16) | ((u32)(SCpnt->cmnd[4]) <<  8) | ((u32)(SCpnt->cmnd[5]));
			blocks = ((u32)(SCpnt->cmnd[7]) << 8) | ((u32)(SCpnt->cmnd[8]));
			if (SCpnt->use_sg)
			{
				// using scatter-gather subsystem, 
				// we first read all data from disk using one read call.
				// after that, we devide the data to the scatterlist.
		//		buffer = kmalloc(blocks*SCpnt->transfersize, GFP_KERNEL);
				malloc_inside_buffer(SCpnt->transfersize*blocks,SCpnt->transfersize);			
			//	PDEBUG("before read data in scatterlist");
				ret = o2media_readData(o2media_pri_data.buffer,block,blocks,devext_addr);
				ptr = o2media_pri_data.buffer;
				sg = (struct scatterlist *) SCpnt->request_buffer;
				mp1_xfer_buf(SCpnt->use_sg,sg,ptr,1,0);
			}
			else
			{
//				PDEBUG("before read data not in scatterlist");
				ret = o2media_readData(SCpnt->request_buffer,block,blocks,devext_addr);
			}
			setscsiresult(SCpnt,ret);
			done(SCpnt);	
//			PDEBUG("readData OK\n");
	//		setcurrentsrb(SCpnt);
		return 0;
		case WRITE_10:
			if(!(SCpnt->request_buffer))
			{
				PDEBUG("request_buffer is NULL when write cmd\n");
				setscsiresult(SCpnt,1);
				done (SCpnt);
				return 0;
			}
			block = ((u32)(SCpnt->cmnd[2]) << 24) | ((u32)(SCpnt->cmnd[3]) << 16) | ((u32)(SCpnt->cmnd[4]) <<  8) | ((u32)(SCpnt->cmnd[5]));
			blocks = ((u32)(SCpnt->cmnd[7]) << 8) | ((u32)(SCpnt->cmnd[8]));
			if (SCpnt->use_sg)
			{
				malloc_inside_buffer(SCpnt->transfersize*blocks,SCpnt->transfersize);			
				ptr = o2media_pri_data.buffer;
				sg = (struct scatterlist *) SCpnt->request_buffer;
				mp1_xfer_buf(SCpnt->use_sg,sg,ptr,0,0);
				ret = o2media_writeData(o2media_pri_data.buffer,block,blocks,devext_addr);
			}
			else
			{
				ret = o2media_writeData(SCpnt->request_buffer,block,blocks,devext_addr);
			}	
			setscsiresult(SCpnt,ret);
			done(SCpnt);	
	//		setcurrentsrb(SCpnt);
		return 0;
		case VERIFY:
		case SEND_DIAGNOSTIC:
		case START_STOP:
		case ALLOW_MEDIUM_REMOVAL:
//			PDEBUG("command return ok"); 
			ret = o2media_allowmediumremoval(devext_addr);
			setscsiresult(SCpnt,ret);
			done (SCpnt);
		return 0;	
		default:
			PDEBUG("command unsupported");
			setscsiresult(SCpnt,1);
			done (SCpnt);
		return 0;
	}
	return 0;
}
static int queuecommand( struct scsi_cmnd *srb , void (*done)(struct scsi_cmnd *))
{
#ifdef SCSI_THREAD
	unsigned long flags;
	if(o2media_pri_data.cur_srb!=NULL)
	{
//		printk("queuecommand return host busy\n");
		return SCSI_MLQUEUE_HOST_BUSY;
	}
	spin_lock_irqsave(&(o2media_pri_data.srb_lock),flags);
	o2media_pri_data.cur_srb = srb;
	spin_unlock_irqrestore(&(o2media_pri_data.srb_lock),flags);
	srb->scsi_done = done;
	spin_lock_irqsave(&(o2media_pri_data.cmd_lock),flags);
	o2media_pri_data.o2media_cmd_comming = 1;
	spin_unlock_irqrestore(&(o2media_pri_data.cmd_lock),flags);
	if(o2media_pri_data.ko2media_cmd_control_pid!=0)
		wake_up_interruptible(&(o2media_pri_data.o2media_cmd_control_event));
	else
	{
		setscsiresult(srb,1);
		done (srb);
	}
	return 0;
#else
	return o2media_cmd_handler(srb,done);
#endif
}
/*
 * this defines our 'host'
 */

// the cmd_per_lun may be bigger than 1?
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
Scsi_Host_Template o2media_host_template = {
#if (KERNEL_VERSION(2,3,99) > LINUX_VERSION_CODE)
//	proc_dir:              &proc_scsi_o2media, /* kernel 2.2 */
#else
// the proc_name can not be NULL, scsi_unregister_host->remove_proc_directory need it.
	proc_name:             "o2media",       /* kernel 2.4 */
#endif
	name:			"o2media host controller",
	info:			host_info,
	detect:			detect,
	release:		release,
	queuecommand:		queuecommand,
	eh_abort_handler:	command_abort,
	eh_device_reset_handler:device_reset,
//	eh_bus_reset_handler:	bus_reset,
//	eh_host_reset_handler:	host_reset,
	can_queue:		1,
	this_id:		-1,
	/*
	if you set SG_ALL, when system request big data(read/write), it will be give you an
	scatterlist, the scatterlist contains the different (address(buffer address),length)
	and the scatter-gather subsystem will be faster than non scatter-gather subsystem.
	if we do not use scatter-gather subsystem, for example, when write an big file to disk,
	system will give you too much request to write, and our device after each write will 
	cost sometime, so the speed becomes slower and slower when the number of request bigger and bigger.
	
	*/
#ifdef SCATTER_GATHER
	sg_tablesize:		SG_ALL,
#else
	sg_tablesize:		0,
#endif
	cmd_per_lun:		1,
	present:		0,
	unchecked_isa_dma:	FALSE,
	use_clustering:		TRUE,	
	/*
	if use_clustering is FALSE, the speed is slow 
	*/
	use_new_eh_code:	TRUE,
	emulated:		TRUE,
};
#else
struct scsi_host_template o2media_host_template = {
	.proc_name              = "o2media",       /* kernel 2.4 */
	.name			= "o2media host controller",
	.info			= host_info,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	.detect			= detect,
#endif
	.release		= release,
	.queuecommand		= queuecommand,
	.eh_abort_handler	= command_abort,
	.eh_device_reset_handler = device_reset,
//	.eh_bus_reset_handler 	= bus_reset,
//	.eh_host_reset_handler	= host_reset,
	.can_queue		= 1,
	.this_id		= -1,
#ifdef SCATTER_GATHER
	.sg_tablesize		= SG_ALL,
#else
	.sg_tablesize		= 0,
#endif
	.cmd_per_lun		= 1,
	.present		= 0,
	.unchecked_isa_dma	= FALSE,
	.use_clustering		= TRUE,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	.use_new_eh_code	= TRUE,
#endif
	.emulated		= TRUE,

};
#endif
